package ParkNPark.client;

import java.io.PrintStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.omg.CORBA.COMM_FAILURE;
import org.omg.CORBA.OBJECT_NOT_EXIST;
import org.omg.CORBA.ORB;
import org.omg.CORBA.SystemException;
import org.omg.CosNaming.NameComponent;
import org.omg.CosNaming.NamingContext;
import org.omg.CosNaming.NamingContextExt;
import org.omg.CosNaming.NamingContextExtHelper;
import org.omg.CosNaming.NamingContextHelper;
import org.omg.CosNaming.NamingContextPackage.CannotProceed;
import org.omg.CosNaming.NamingContextPackage.InvalidName;
import org.omg.CosNaming.NamingContextPackage.NotFound;

import ParkNPark.common.InputEater;
import ParkNPark.interfaces.AlreadyInLotException;
import ParkNPark.interfaces.AtBottomLevelException;
import ParkNPark.interfaces.AtTopLevelException;
import ParkNPark.interfaces.ClientManager;
import ParkNPark.interfaces.ClientManagerFactory;
import ParkNPark.interfaces.ClientManagerFactoryHelper;
import ParkNPark.interfaces.ClientManagerOperations;
import ParkNPark.interfaces.InvalidClientException;
import ParkNPark.interfaces.LotFullException;
import ParkNPark.interfaces.LotNotFoundException;
import ParkNPark.interfaces.NotInLotException;
import ParkNPark.interfaces.NotOnExitLevelException;
import ParkNPark.interfaces.ServiceUnavailableException;
import ParkNPark.interfaces._ClientManagerStub;

/**
 * Fault tolerant client manager that wraps a remote ClientManager and rolls
 * over to a new primary server in the presence of failures.
 * Note that this class is not multi-thread safe. Multiple threads should
 * synchronize access or wrap this class with a thread-safe delegate.
 */
public class FaultTolerantClientManager implements ClientManagerOperations
{
	/** Our naming service's ParkNPark context, which contains server registrations */
	protected NamingContext parkNParkContext = null;
	
	/** The primary client manager */
	protected ClientManager primaryClientManager = null;

	/** Our name server instance */
	protected NamingContextExt nameServer;
	
	/** The latest sequence number from the client */
	protected int seq;
	
	/** The server-generated client ID */
	protected int clientID;

	/** Our object request broker */
	protected ORB orb;

	/**
	 * Creates a new fault-tolerant client manager that automatically fails over
	 * to new primary servers on method call failures
	 * @param orb The object request broker instance to use
	 * @throws ServiceUnavailableException Thrown when a primary server could not be
	 * located or contacted
	 */
	public FaultTolerantClientManager(ORB orb) throws ServiceUnavailableException {
		this.orb = orb;
		
		// If we could not get a primary server, then throw a ServiceUnavailableException
		if (!getPrimaryServer())
			throw new ServiceUnavailableException("Could not contact a primary server");
	}
	
	/**
	 * Return value or an exception along with the return value or the exception
	 * object. The exceptionThrown flag indicates whether this is an exception
	 * or a normal return
	 */
	protected class ReturnValueOrException
	{
		/** The return value or exception object */
		public Object retVal = null;
		
		/** When true, this instance is an exception; otherwise, this is a normal return value */
		public boolean exceptionThrown = false;
		
		/**
		 * Creates a new ReturnValueOrException using the given return value or exception object
		 * and whether or not that object is a return value or an exception
		 * @param retVal The return value or exception object
		 * @param exceptionThrown When true, retVal is an exception; when false, it's a return value
		 */
		public ReturnValueOrException(Object retVal, boolean exceptionThrown) {
			this.retVal = retVal;
			this.exceptionThrown = exceptionThrown;
		}
	}
	
	/**
	 * Calls the given method on the server with the given arguments and returns
	 * a ReturnValueOrException object to indicate a normal or exceptional result
	 * and contain the return value or the exception object. If server communication
	 * or service unavailability exceptions arise, this method will automatically
	 * fail over to the new primary after that becomes available. Server communication
	 * and service unavailability exceptions are only returned if no primary
	 * server is registered
	 * @param method The method to call on our client manager on the primary server
	 * @param arguments The arguments to pass to the given method's parameters
	 * @return A ReturnValueOrException indicating a normal or exceptional result
	 */
	protected ReturnValueOrException callServer(Method method, Object... arguments)	    
	{
		// Try until the method works or is hopeless
		int attempts = 0;
		Object retVal;
		while (true)
		{
			// Disable CORBA-spewed stack traces (restore immediately after calling method.invoke)
			PrintStream err = System.err;
			System.setErr(inputEater);
			
			// Attempt to call the method
			try {
				retVal = method.invoke(primaryClientManager, arguments);
			    System.setErr(err);
			    return new ReturnValueOrException(retVal, false);
			} catch (InvocationTargetException e) {
			    System.setErr(err);

			    // This is an exception
				retVal = e.getCause();

				// Check for COMM_FAILURE, OBJECT_NOT_EXIST, and ServiceUnavailable exceptions
				if (retVal instanceof COMM_FAILURE || retVal instanceof OBJECT_NOT_EXIST ||
					retVal instanceof ServiceUnavailableException) {
					if (retVal instanceof ServiceUnavailableException)
						System.out.println("Server is not operating properly; getting another server...");
					else if (retVal instanceof COMM_FAILURE)
						System.out.println("Problem while communicating with the server; getting another server...");
					else if (retVal instanceof OBJECT_NOT_EXIST)
						System.out.println("Problem while communicating with the server; getting another server...");
					
					// We need a new primary!
					if (!getPrimaryServer())
					{
						// We could not get a primary server; inform the client
						System.out.println("No servers are up");
						return new ReturnValueOrException(retVal, true);
					} else {
						attempts++;
						
						// If this is our second or subsequent attempt, then wait for five seconds first
						// (the replication manager needs time to see that the primary server has
						// database communication problems, too)
						if (attempts >= 2) {
							try {
								Thread.sleep(5000);
							} catch (InterruptedException f) { }
						}
					}
				}

				// Otherwise, normal exception was thrown by this method invocation; give that to our caller
				else
					return new ReturnValueOrException(retVal, true);
			} catch (IllegalAccessException e) {
			    System.setErr(err);

			    // A reflection API exception occurred; throw this as a RuntimeException to the caller
				// (this thread does not have sufficient access rights to use this part of the
				// reflection API)
				System.out.println("Client is not permitted to use the Java Reflection API, which is required for normal operation; exiting");
				throw new RuntimeException(e);
			}
		}
	}

	/**
	 * Refreshes the primary server that is registered in the name service
	 * @return True if the primary server reference is refreshed or false if it could not
	 * be refreshed
	 */
	protected boolean getPrimaryServer()
	{
		// If we don't have our ParkNPark naming context, then get it from the server now.
		// If that does not work, then return false now
		if (parkNParkContext == null && !refreshParkNParkContext())
			return false;
		
		// Attempt to get the new primary client manager
		ClientManagerFactory factory;
		int attempts = 0;
		boolean debugMessages = false;
		while (true)
		{
			// Look up the primary server
			try {
				if (debugMessages) System.out.println("Looking up the primary server from the name service");
				NameComponent[] serverName = nameServer.to_name("ParkNParkPrimary");
				factory = ClientManagerFactoryHelper.narrow(parkNParkContext.resolve(serverName));
			} catch (InvalidName e) {
				if (debugMessages) System.out.println("Unexpected exception: " + e.toString());
				return false;
			} catch (NotFound e)
			{
				// No servers registered
				if (debugMessages) System.out.println("No servers are active at this time");
				return false;
			} catch (CannotProceed e) {
				if (debugMessages) System.out.println("Unexpected exception: " + e.toString());
				return false;
			} catch (COMM_FAILURE e)
			{
				// Refresh the parkNPark context and, if that fails, notify the user
				if (debugMessages) System.out.println("Problem communicating with the name service; retrying...");
				if (!refreshParkNParkContext()) {
					System.out.println("Name service is not accessible");
					return false;
				}
				
				// Start from the beginning
				continue;
			} catch (OBJECT_NOT_EXIST e)
			{
				// Refresh the parkNPark context and, if that fails, notify the user
				if (debugMessages) System.out.println("Name service object disappeared; retrying...");
				if (!refreshParkNParkContext()) {
					System.out.println("Name service is not accessible");
					return false;
				}
				
				// Start from the beginning
				continue;
			} catch (SystemException e) {
				System.out.println("Communication problem while communicating with the name server");
				return false;
			} catch (Exception e) {
				System.out.println("Miscellaneous exception (possibly a program bug?): " + e.getClass().getName());
				return false;
			}
			
			// Get our client manager object
			ClientManager newClientManager = null;
			PrintStream err = System.err;
			try {
				// If we already had a primary client manager, then ask for our existing client manager; otherwise,
				// create a new instance
				if (debugMessages) System.out.println("Obtaining our client manager");
				System.setErr(inputEater);
				if (primaryClientManager == null) {
				    newClientManager = factory.getClientManager();
				} else
					newClientManager = factory.getExistingClientManager(clientID);
				System.setErr(err);
				if (debugMessages) System.out.println("Client manager obtained");
			} catch (ServiceUnavailableException e) {
				System.setErr(err);
				System.out.println("Server is not functioning properly; trying another server");
			} catch (InvalidClientException e) {
				System.setErr(err);
				
				// Our client's state no longer is in the system!
				System.out.println("Server does not remember us; we cannot continue");
				return false;
			} catch (COMM_FAILURE e) {
				System.setErr(err);
				
				// Communication exception while communicating with the server; try to get the primary again
				System.out.println("Communication problem while communicating with the server; trying another server...");
			} catch (OBJECT_NOT_EXIST e) {
				System.setErr(err);
				System.out.println("Communication problem while communicating with the server; trying another server...");
			} catch (SystemException e) {
				System.setErr(err);
				System.out.println("Unhandled CORBA exception while communicating with the client manager factory: " + e.getClass().getName());
				return false;
			} catch (Exception e) {
				System.setErr(err);
				System.out.println("Miscellaneous exception (possibly a program bug?): " + e.getClass().getName());
				return false;
			}
			
			// Use this client manager if we are supposed to
			if (newClientManager != null)
			{
				try {
					// Get the client ID from the server
					int serverClientID = newClientManager.getClientID();
					
					// Set the new client manager synchronously
					System.out.println("Server successfully obtained");
					synchronized (this) {
						primaryClientManager = newClientManager;
						clientID = serverClientID;
					}
					
					// We're done :-)
					return true;
				} catch (COMM_FAILURE e) {
					// Communication exception while communicating with the server; try to get the primary again
					System.out.println("Communication problem while communicating with the server; trying another server...");
				} catch (OBJECT_NOT_EXIST e) {
					System.out.println("Object does not exist on the server; trying to get the primary again: " + e.getClass().getName());
				} catch (SystemException e) {
					System.out.println("Unhandled CORBA exception while communicating with the client manager factory: " + e.getClass().getName());
					return false;
				} catch (Exception e) {
					System.out.println("Miscellaneous exception (possibly a program bug?): " + e.getClass().getName());
					return false;
				}
			}

			// We need to get our primary again
			attempts++;
			
			// If this is our second or subsequent attempt, then wait for five seconds first
			if (attempts >= 2) {
				try {
					System.out.println("Waiting for a server to become available...");
					Thread.sleep(5000);
				} catch (InterruptedException f) { }
			}
		}
	}
	
	/**
	 * Refreshes the parkNParkContext name service object that we use to get server
	 * bindings on. The name service is contacted from scratch in this method
	 * @return True if the parkNParkContext was refreshed or false if it was not
	 */
	protected boolean refreshParkNParkContext()
	{
		// Fetch the root naming context
		PrintStream err = System.err;
		try {
			System.setErr(inputEater);
		    nameServer = NamingContextExtHelper.narrow(orb.resolve_initial_references("NameService"));
		    System.setErr(err);
		} catch (Exception e) {
		    System.setErr(err);
			System.out.println("Could not contact the naming service; please ensure that it is running.");
			parkNParkContext = null;
			return false;
		}

		// Get the proper context object for ParkNPark
		System.out.println("Fetching the ParkNPark name service context");
		NameComponent parkNParkContextName[];
		try {
		    parkNParkContextName = nameServer.to_name("ParkNPark");
		} catch (Exception e) {
			System.out.println("Communication problem while communicating with the name server");
			parkNParkContext = null;
			return false;
		}
		try {
			parkNParkContext = NamingContextHelper.narrow(nameServer.resolve(parkNParkContextName));
		} catch (NotFound e) {
			System.out.println("ParkNPark name service context not found; is the ParkNPark server up?");
			parkNParkContext = null;
			return false;
		} catch (SystemException e) {
			System.out.println("Communication problem while communicating with the name server");
			parkNParkContext = null;
			return false;
		} catch (Exception e) {
			System.out.println("Miscellaneous exception (possibly a program bug?): " + e.getClass().getName());
			parkNParkContext = null;
			return false;
		}
		
		// We were able to get it :-)
		return true;
	}
	
	protected static final Method getClientID;
	protected static final Method enterLot;
	protected static final Method exitLot;
	protected static final Method getOtherLotAvailability;
	protected static final Method getLots;
	protected static final Method moveUpLevel;
	protected static final Method moveDownLevel;
	protected static final Method getCurrentLevel;
	protected static final Method getMaxLevel;
	protected static final Method getMinLevel;
	protected static final Method closeClientManager;

	/** Initialize our cached Method instances */
	static {
		try {
		   Class noParameters[] = new Class[0];
		   Class oneIntParameter[] = new Class[] { int.class };
		   Class twoIntParameters[] = new Class[] { int.class, int.class };
		   getClientID = _ClientManagerStub.class.getMethod("getClientID", noParameters);
		   enterLot = _ClientManagerStub.class.getMethod("enterLot", twoIntParameters);
		   exitLot = _ClientManagerStub.class.getMethod("exitLot", oneIntParameter);
		   getOtherLotAvailability = _ClientManagerStub.class.getMethod("getOtherLotAvailability", oneIntParameter);
		   getLots = _ClientManagerStub.class.getMethod("getLots", noParameters);
		   moveUpLevel = _ClientManagerStub.class.getMethod("moveUpLevel", oneIntParameter);
		   moveDownLevel = _ClientManagerStub.class.getMethod("moveDownLevel", oneIntParameter);
		   getCurrentLevel = _ClientManagerStub.class.getMethod("getCurrentLevel", noParameters);
		   getMaxLevel = _ClientManagerStub.class.getMethod("getMaxLevel", noParameters);
		   getMinLevel = _ClientManagerStub.class.getMethod("getMinLevel", noParameters);
		   closeClientManager = _ClientManagerStub.class.getMethod("closeClientManager", noParameters);
		} catch (NoSuchMethodException e) {
			throw new RuntimeException(e);
		}
	}
	
	/** Message for unexpected exceptions that arise during remote method calls (non-IDL and non-CORBA exceptions) */
	protected final static String unexpectedException = "Unexpected exception thrown";
	
    /**
     * Returns the client's ID, which can be used in the client manager factory's
     * getExistingClientManager() method
     * @return The client's ID
     */
	public int getClientID() {
		ReturnValueOrException retVal = callServer(getClientID);
		if (retVal.exceptionThrown) {
			if (retVal.retVal instanceof RuntimeException)
				throw (RuntimeException) retVal.retVal;
			else
			    throw new RuntimeException(unexpectedException, (Throwable) retVal.retVal);
		} else
			return ((Integer) retVal.retVal).intValue();
	}

	/**
     * Moves the client's car into the lot with the given lot number and returns an array of level
     * numbers that have available spaces
     * @param seq The latest sequence number of the client
     * @param lot The lot number to enter
     * @throws AlreadyInLotException Thrown when the client's car is already in a lot
     * @throws LotNotFoundException Thrown if the given lot number is not known to the system
     * @throws LotFullException Thrown if the given lot is full
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the car from entering the lot
     * @return An array of level numbers that have available spaces
     * [Category: Baseline; Requirements: 1, 2, and 3]
     */
	public int[] enterLot(int seq, int lot) throws AlreadyInLotException, LotNotFoundException, LotFullException, ServiceUnavailableException {
		this.seq = seq;
		ReturnValueOrException retVal = callServer(enterLot, seq, lot);
		if (retVal.exceptionThrown) {
			if (retVal.retVal instanceof AlreadyInLotException)
				throw (AlreadyInLotException) retVal.retVal;
			else if (retVal.retVal instanceof LotNotFoundException)
				throw (LotNotFoundException) retVal.retVal;
			else if (retVal.retVal instanceof LotFullException)
				throw (LotFullException) retVal.retVal;
			else if (retVal.retVal instanceof ServiceUnavailableException)
				throw (ServiceUnavailableException) retVal.retVal;
			else if (retVal.retVal instanceof RuntimeException)
				throw (RuntimeException) retVal.retVal;
			else
			    throw new RuntimeException(unexpectedException, (Throwable) retVal.retVal);
		} else
		    return (int[]) retVal.retVal;
	}

	/**
     * Removes the client's car from the lot that it is currently in
     * @param seq The latest sequence number of the client
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws NotOnExitLevelException Thrown if the car is in a lot but is not on a permitted
     * exit level
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the car from exiting the lot
     * [Category: Baseline; Requirement: 7]
     */
	public void exitLot(int seq) throws NotInLotException, NotOnExitLevelException, ServiceUnavailableException {
		this.seq = seq;
		ReturnValueOrException retVal = callServer(exitLot, seq);
		if (retVal.exceptionThrown) {
			if (retVal.retVal instanceof NotInLotException)
				throw (NotInLotException) retVal.retVal;
			else if (retVal.retVal instanceof NotOnExitLevelException)
				throw (NotOnExitLevelException) retVal.retVal;
			else if (retVal.retVal instanceof ServiceUnavailableException)
				throw (ServiceUnavailableException) retVal.retVal;
			else if (retVal.retVal instanceof RuntimeException)
				throw (RuntimeException) retVal.retVal;
			else
			    throw new RuntimeException(unexpectedException, (Throwable) retVal.retVal);
		}
	}

	/**
     * Returns an array of other lots that have availability, sorted by lot distance such that
     * closer lots are listed first
     * @param lot The lot to get lot distances from
     * @throws LotNotFoundException Thrown if the given lot number is not known to the system
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if
     * some other reason prevents the system from discovering the availability of other lots
     * @return An array of other lots that have availability
     * [Category: Baseline; Requirement: 4]
     */
	public int[] getOtherLotAvailability(int lot) throws LotNotFoundException, ServiceUnavailableException {
		ReturnValueOrException retVal = callServer(getOtherLotAvailability, lot);
		if (retVal.exceptionThrown) {
			if (retVal.retVal instanceof LotNotFoundException)
				throw (LotNotFoundException) retVal.retVal;
			else if (retVal.retVal instanceof ServiceUnavailableException)
				throw (ServiceUnavailableException) retVal.retVal;
			else if (retVal.retVal instanceof RuntimeException)
				throw (RuntimeException) retVal.retVal;
			else
			    throw new RuntimeException(unexpectedException, (Throwable) retVal.retVal);
		} else
		    return (int[]) retVal.retVal;
	}

	/**
     * Returns an array of valid lot numbers in the system, sorted by the lot number in ascending order
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the system from discovering the its defined lots
     * @return An array of valid lot numbers in the system
     * [Category: Baseline; Requirement: 12]
     */
	public int[] getLots() throws ServiceUnavailableException {
		ReturnValueOrException retVal = callServer(getLots);
		if (retVal.exceptionThrown) {
			if (retVal.retVal instanceof ServiceUnavailableException)
				throw (ServiceUnavailableException) retVal.retVal;
			else if (retVal.retVal instanceof RuntimeException)
				throw (RuntimeException) retVal.retVal;
			else
			    throw new RuntimeException(unexpectedException, (Throwable) retVal.retVal);
		} else
		    return (int[]) retVal.retVal; 
	}

	/**
     * Moves the car from its present level to the level above it
     * @param seq The latest sequence number of the client
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws AtTopLevelException Thrown if the car is already on the highest level
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the car from moving to the next highest level
     * @return The level number that the client's car is now on
     * [Category: Baseline; Requirement: 6]
     */
	public int moveUpLevel(int seq) throws NotInLotException, AtTopLevelException, ServiceUnavailableException {
		this.seq = seq;
		ReturnValueOrException retVal = callServer(moveUpLevel, seq);
		if (retVal.exceptionThrown) {
			if (retVal.retVal instanceof NotInLotException)
				throw (NotInLotException) retVal.retVal;
			else if (retVal.retVal instanceof AtTopLevelException)
				throw (AtTopLevelException) retVal.retVal;
			else if (retVal.retVal instanceof ServiceUnavailableException)
				throw (ServiceUnavailableException) retVal.retVal;
			else if (retVal.retVal instanceof RuntimeException)
				throw (RuntimeException) retVal.retVal;
			else
			    throw new RuntimeException(unexpectedException, (Throwable) retVal.retVal);
		} else
		    return ((Integer) retVal.retVal).intValue();
	}

	/**
     * Moves the car from its present level to the level beneath it
     * @param seq The latest sequence number of the client
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws AtTopLevelException Thrown if the car is already on the lowest level
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some other reason prevents the car from moving to the lower level
     * @return The level number that the client's car is now on
     * [Category: Baseline; Requirement: 6]
     */
	public int moveDownLevel(int seq) throws NotInLotException, AtBottomLevelException, ServiceUnavailableException {
		this.seq = seq;
		ReturnValueOrException retVal = callServer(moveDownLevel, seq);
		if (retVal.exceptionThrown) {
			if (retVal.retVal instanceof NotInLotException)
				throw (NotInLotException) retVal.retVal;
			else if (retVal.retVal instanceof AtBottomLevelException)
				throw (AtBottomLevelException) retVal.retVal;
			else if (retVal.retVal instanceof ServiceUnavailableException)
				throw (ServiceUnavailableException) retVal.retVal;
			else if (retVal.retVal instanceof RuntimeException)
				throw (RuntimeException) retVal.retVal;
			else
			    throw new RuntimeException(unexpectedException, (Throwable) retVal.retVal);
		} else
		    return ((Integer) retVal.retVal).intValue();
	}

	/**
     * Returns the car's current level number
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the system from returning the car's current level
     * @return The car's current level number
     * [Category: Baseline; Requirement: 6]
     */
	public int getCurrentLevel() throws NotInLotException, ServiceUnavailableException {
		ReturnValueOrException retVal = callServer(getCurrentLevel);
		if (retVal.exceptionThrown) {
			if (retVal.retVal instanceof NotInLotException)
				throw (NotInLotException) retVal.retVal;
			else if (retVal.retVal instanceof ServiceUnavailableException)
				throw (ServiceUnavailableException) retVal.retVal;
			else if (retVal.retVal instanceof RuntimeException)
				throw (RuntimeException) retVal.retVal;
			else
			    throw new RuntimeException(unexpectedException, (Throwable) retVal.retVal);
		} else
		    return ((Integer) retVal.retVal).intValue();
	}

	/**
     * Returns the top level number of the car's current parking lot
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the system from returning the current lot's highest level
     * @return The top level number of the car's current parking lot
     * [Category: Baseline; Requirement: 6]
     */
	public int getMaxLevel() throws NotInLotException, ServiceUnavailableException {
		ReturnValueOrException retVal = callServer(getMaxLevel);
		if (retVal.exceptionThrown) {
			if (retVal.retVal instanceof NotInLotException)
				throw (NotInLotException) retVal.retVal;
			else if (retVal.retVal instanceof ServiceUnavailableException)
				throw (ServiceUnavailableException) retVal.retVal;
			else if (retVal.retVal instanceof RuntimeException)
				throw (RuntimeException) retVal.retVal;
			else
			    throw new RuntimeException(unexpectedException, (Throwable) retVal.retVal);
		} else
		    return ((Integer) retVal.retVal).intValue();
	}

	/**
     * Returns the bottom level number of the car's current parking lot
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the system from returning the current lot's lowest level
     * @return The bottom level number of the car's current parking lot
     * [Category: Baseline; Requirement: 6]
     */
	public int getMinLevel() throws NotInLotException, ServiceUnavailableException {
		ReturnValueOrException retVal = callServer(getMinLevel);
		if (retVal.exceptionThrown) {
			if (retVal.retVal instanceof NotInLotException)
				throw (NotInLotException) retVal.retVal;
			else if (retVal.retVal instanceof ServiceUnavailableException)
				throw (ServiceUnavailableException) retVal.retVal;
			else if (retVal.retVal instanceof RuntimeException)
				throw (RuntimeException) retVal.retVal;
			else
			    throw new RuntimeException(unexpectedException, (Throwable) retVal.retVal);
		} else
		    return ((Integer) retVal.retVal).intValue();
	}

	/**
     * Closes the client manager and frees server resources associated with it, including the client
     * manager's activation in the server's CORBA portable object adapter
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the system from closing the client manager. The client manager remains
     * open if this exception gets thrown
     */
	public void closeClientManager() throws ServiceUnavailableException {
		ReturnValueOrException retVal = callServer(closeClientManager);
		if (retVal.exceptionThrown) {
			if (retVal.retVal instanceof ServiceUnavailableException)
				throw (ServiceUnavailableException) retVal.retVal;
			else if (retVal.retVal instanceof RuntimeException)
				throw (RuntimeException) retVal.retVal;
			else
			    throw new RuntimeException(unexpectedException, (Throwable) retVal.retVal);
		}
	}
	
	/** Shared static instance of the input eater output stream wrapped in a PrintStream */
	protected static PrintStream inputEater = new PrintStream(new InputEater());
}
