package ParkNPark.middletier;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;

import ParkNPark.interfaces.AlreadyInLotException;
import ParkNPark.interfaces.AtBottomLevelException;
import ParkNPark.interfaces.AtTopLevelException;
import ParkNPark.interfaces.ClientManagerPOA;
import ParkNPark.interfaces.InvalidClientException;
import ParkNPark.interfaces.LotFullException;
import ParkNPark.interfaces.LotNotFoundException;
import ParkNPark.interfaces.NotInLotException;
import ParkNPark.interfaces.NotOnExitLevelException;
import ParkNPark.interfaces.ServiceUnavailableException;

/**
 * Server object that will manage communication between the client
 * and the database.
 */
public class ClientManagerImpl extends ClientManagerPOA
{
	/** Maximum number of attempts when trying to execute commends on database */
	private static final int MAX_ATTEMPTS = 3;
	
	/** Flags used to print out the status of lots and levels */
	private static final boolean PRINT_LOTS_STATUS = true;
	private static final boolean PRINT_LEVELS_STATUS = true;
	
	/** Client's ID used for fault-tolerance */
	private int clientID;
	
	/** Client's current operation sequence number used for fault-tolerance */
	private int seq;
	
	/** Lot that client is in */
	private int lot;
	
	/** Level that client is in */
	private int level;
	
	/** Connection to the database */
	private Connection conn;
	
	/** Statements used for duplicate message detection */
	PreparedStatement checkDuplicate;
	PreparedStatement updateSequence;
	
	/**
	 * Create a client manager for an existing client that will hold
	 * information specific to the client.
	 * @param clientID
	 * @param seq
	 * @throws ServiceUnavailableException Thrown when a database exception prevented the client
	 * manager from being created successfully
	 */
	public ClientManagerImpl(int clientID) throws ServiceUnavailableException, InvalidClientException
	{
		// Setup database connection
		setupDatabaseConn();
		
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{	try
			{	// Check if valid client id
				PreparedStatement ps = conn.prepareStatement("SELECT Seq, LotID, Level FROM Client WHERE ClientID = ?");
				ps.setInt(1, clientID);
				ResultSet rs = ps.executeQuery();
				
				if (rs.next())
				{	// Store client's id and sequence number
					this.clientID = clientID;
					this.seq = rs.getInt("Seq");
					this.lot = rs.getInt("LotID");
					if (rs.wasNull())
						this.lot = -1;
					this.level = rs.getInt("Level");
					if (rs.wasNull())
						this.level = -1;
					rs.close();
					
					return;
				}
				else
				{	throw new InvalidClientException();
				}
			}
			catch (SQLException e)
			{	try
				{	conn.rollback();
					numTries++;
				}
				catch (SQLException se)
				{	throw new ServiceUnavailableException("Could not rollback when creating existing client.");
				}
			}
		}
	
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not create existing client.");
	}
	
	/**
	 * Create a client manager for a new client that will hold 
	 * information specific to the client.
	 * @throws ServiceUnavailableException Thrown when a database exception prevented the client
	 * manager from being created successfully
	 */
	public ClientManagerImpl() throws ServiceUnavailableException
	{
		// Setup database connection
		setupDatabaseConn();
		
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{	try
			{	// Create a new client id
				Statement stmt = conn.createStatement();
				stmt.executeUpdate("INSERT INTO Client (Seq, LotID, Level) VALUES(0, null, null)");
				ResultSet rs = stmt.executeQuery("SELECT last_insert_id() ClientID");
				if (rs.next())
				{	this.clientID = rs.getInt("ClientID");
					this.seq = 0;
					this.lot = -1;
					this.level = -1;
					conn.commit();
					return;
				}
				else
				{	numTries++;
				}
			}
			catch (SQLException e)
			{	try
				{	conn.rollback();
					numTries++;
				}
				catch (SQLException se)
				{	throw new ServiceUnavailableException("Could not rollback when creating "
						+ "new client id");
				}
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not create new client id.");
	}
	
	/**
	 * Set up the database connection
	 * @throws ServiceUnavailableException Thrown if a problem came up while connecting to the database server
	 * or setting it up after connecting
	 */
	private void setupDatabaseConn() throws ServiceUnavailableException
	{
		// Retry on failed attempts
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{	try
			{	// Set database configuration
				String jdbcDriver = "com.mysql.jdbc.Driver";
				String jdbcURL = "jdbc:mysql://172.19.132.120:3306/ece749_team6";
				String jdbcUsername = "team6";
				String jdbcPassword = "procrastinators";
				
				// Load the JDBC driver
				try 
				{	Class.forName(jdbcDriver);
				} 
				catch (ClassNotFoundException e) 
				{	System.err.println("Could not load JDBC driver; ensure that it is within the classpath");
					throw new RuntimeException(e);
				}
				
				// Connect to the database
				conn = DriverManager.getConnection(jdbcURL, jdbcUsername, jdbcPassword);
				
				// Turn off auto-commit, enabling transaction control
				conn.setAutoCommit(false);
				
				// Set up prepared statements
				this.checkDuplicate = conn.prepareStatement("SELECT Seq FROM Client WHERE ClientID = ? AND Seq >= ?");
				this.updateSequence = conn.prepareStatement(
						"UPDATE Client SET Seq = ?, LotID = ?, Level = ? WHERE ClientID = ?");
				
				return;
			} 
			catch (SQLException e) 
			{	numTries++;
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not create connection.");
	}

	/**
     * Closes the client manager and frees server resources associated with it, including the client
     * manager's activation in the server's CORBA portable object adapter
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the system from closing the client manager. The client manager remains
     * open if this exception gets thrown
     */
	public void closeClientManager() throws ServiceUnavailableException 
	{
		// No need to close connection if it's already closed
		if (conn == null)
			return;
		
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{	try
			{	conn.close();
				conn = null;
				
				// Stop serving this object now that it is closed
				try 
				{	_poa().deactivate_object(_poa().servant_to_id(this));
				} 
				catch (Exception e) 
				{	System.err.println("CORBA exception while closing the client manager: " + e.getClass().getName());
				}
				return;
			}
			catch(SQLException e)
			{	numTries++;
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not close database connection.");
	}

    /**
     * Moves the client's car into the lot with the given lot number and returns an array of level
     * numbers that have available spaces
     * @param seq The latest sequence number of the client
     * @param lot The lot number to enter
     * @throws AlreadyInLotException Thrown when the client's car is already in a lot
     * @throws LotNotFoundException Thrown if the given lot number is not known to the system
     * @throws LotFullException Thrown if the given lot is full
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the car from entering the lot
     * @return An array of level numbers that have available spaces
     */
	public int[] enterLot(int seq, int lot) throws AlreadyInLotException, LotNotFoundException, LotFullException, 
		ServiceUnavailableException 
	{
		printLotsStatus("Before entering lot", seq);
		
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{
			try
			{	// Check if lot is valid
				PreparedStatement lotInfo = conn.prepareStatement("SELECT CarCount, Capacity FROM Lot WHERE ID = ?");
				lotInfo.setInt(1, lot);
				ResultSet rs = lotInfo.executeQuery();
				
				if (rs.next())
				{	// Check if request already completed
					if (alreadyCompleted(seq))
					{	System.out.println("Request already completed (in enter lot)");
					}
					else
					{	// If the client is already in a lot, then throw an AlreadyInLotException now
						if (this.lot != -1)
							throw new AlreadyInLotException(this.lot);
					
						// Check if lot is full
						int carCount = rs.getInt("CarCount");
						int capacity = rs.getInt("Capacity");
						rs.close();
					
						if (carCount < capacity)
						{
							// Increment the lot's car count
							PreparedStatement enterLot = conn.prepareStatement(
								"UPDATE Lot SET CarCount = CarCount + 1 WHERE ID = ?");
							enterLot.setInt(1, lot);
							if (enterLot.executeUpdate() != 1) {
								conn.rollback();
								throw new ServiceUnavailableException("Update on Lot table failed unexpectedly");
							}
	
							// Increment the entry level's car count
							PreparedStatement enterLevel = conn.prepareStatement(
								"UPDATE Level SET CarCount = CarCount + 1 WHERE LotID = ? AND Level = ?");
							enterLevel.setInt(1, lot);
							enterLevel.setInt(2, 1);
							if (enterLevel.executeUpdate() != 1) {
								conn.rollback();
								throw new ServiceUnavailableException("Could not add car to entry level; perhaps the entry level does not exist?");
							}
						}
						else
						{	throw new LotFullException();
						}
					}
						
					// Get list of available levels
					PreparedStatement availableLevels = conn.prepareStatement(
						"SELECT Level FROM Level WHERE LotID = ? AND CarCount < Capacity ORDER BY Level");
					availableLevels.setInt(1, lot);
					rs = availableLevels.executeQuery();
					
					rs.last();
					int[] availLevels = new int[rs.getRow()];
					int i = 0;
					rs.beforeFirst();
					while (rs.next())
						availLevels[i++] = rs.getInt("Level");
					rs.close();
					
					// Update client data and commit the changes
					updateClientData(seq, lot, 1);
					conn.commit();
					
					// With the changes committed, update our cached values
					this.seq = seq;
					this.lot = lot;
					this.level = 1;
					
					printLotsStatus("After entering lot", seq);
					printLevelsStatus("After entering lot", seq);
					
					return availLevels;
				}
				else
				{	throw new LotNotFoundException();
				}
			}
			catch (SQLException e)
			{	try
				{	conn.rollback();
					numTries++;
				}
				catch (SQLException se)
				{	throw new ServiceUnavailableException("Could not rollback when entering lot.");
				}
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not enter lot.");
	}

    /**
     * Removes the client's car from the lot that it is currently in
     * @param seq The latest sequence number of the client
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws NotOnExitLevelException Thrown if the car is in a lot but is not on a permitted
     * exit level
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the car from exiting the lot
     */
	public void exitLot(int seq) throws NotInLotException, NotOnExitLevelException, ServiceUnavailableException
	{
		printLotsStatus("Before exiting lot", seq);
		printLevelsStatus("Before exiting lot", seq);
		
		// Check if request already completed
		if (alreadyCompleted(seq))
		{	System.out.println("Request already completed (in exit lot)");
			printLotsStatus("After exiting lot", seq);
			printLevelsStatus("After exiting lot", seq);
			return;
		}
		
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{
			try
			{   // If we are not in a lot, then throw an exception
				if (lot == -1)
					throw new NotInLotException();
				
				// If we are not on an exit level, then throw an exception
				if (level != 1)
					throw new NotOnExitLevelException(lot, level);
				
				// Get car count of lot
				PreparedStatement lotCarCount = conn.prepareStatement("SELECT CarCount FROM Lot WHERE ID = ?");
				lotCarCount.setInt(1, this.lot);
				ResultSet rs = lotCarCount.executeQuery();
				rs.next();
				int count = rs.getInt("CarCount");
				rs.close();
				
				// If lot car count is not zero, decrement count
				if (count > 0)
				{	PreparedStatement exitLot = conn.prepareStatement(
						"UPDATE Lot SET CarCount = CarCount - 1 WHERE ID = ?");
					exitLot.setInt(1, lot);
					exitLot.execute();
				}
				
				// Get car count of level
				PreparedStatement levelCarCount = conn.prepareStatement("SELECT CarCount FROM Level WHERE LotID = ? AND Level = ?");
				levelCarCount.setInt(1, lot);
				levelCarCount.setInt(2, 1);
				rs = levelCarCount.executeQuery();
				rs.next();
				count = rs.getInt("CarCount");
				rs.close();
				
				// If level car count is not zero, decrement count
				if (count > 0)
				{	PreparedStatement exitLevel = conn.prepareStatement(
						"UPDATE Level SET CarCount = CarCount - 1 WHERE LotID = ? AND Level = ?");
					exitLevel.setInt(1, lot);
					exitLevel.setInt(2, 1);
					exitLevel.execute();
				}
				
				printLotsStatus("After exiting lot", seq);
				printLevelsStatus("After exiting lot", seq);
				
				// Update client data and commit the changes
				updateClientData(seq, -1, -1);
				conn.commit();
				
				// With the changes committed, update our cached values
				this.seq = seq;
				this.lot = -1;
				this.level = -1;
				
				return;
			}
			catch (SQLException e)
			{	try
				{	conn.rollback();
					numTries++;
				}
				catch (SQLException se)
				{	throw new ServiceUnavailableException("Could not rollback when exiting lock.");
				}
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not exit lot.");	
	}

    /**
     * Returns the client's ID, which can be used in the client manager factory's
     * getExistingClientManager() method
     * @return The client's ID
     */
	public int getClientID() 
	{
		return this.clientID;
	}
	
	/**
	 * Returns the client's sequence number, which can be used if the client loses
	 * track of it
	 * @return
	 */
	public int getSeqNum()
	{
		return this.seq;
	}

	/**
     * Returns the car's current level number
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the system from returning the car's current level
     * @return The car's current level number
     */
	public int getCurrentLevel() throws NotInLotException, ServiceUnavailableException
	{
		// If we are not in a lot, then throw an exception
		if (lot == -1)
			throw new NotInLotException();
		
		// Return the current level
		return this.level;
	}

	/**
     * Returns an array of valid lot numbers in the system, sorted by the lot number in ascending order
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the system from discovering the its defined lots
     * @return An array of valid lot numbers in the system
     */
	public int[] getLots() throws ServiceUnavailableException
	{
		printLotsStatus("When getting lots", seq);
		
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{
			try
			{	// Get list of available lots
				PreparedStatement availableLots = conn.prepareStatement(
					"SELECT ID FROM Lot ORDER BY ID");
				ResultSet rs = availableLots.executeQuery();
				rs.last();				
				int[] lots = new int[rs.getRow()];
				rs.beforeFirst();
				
				int i = 0;
				while (rs.next())
					lots[i++] = rs.getInt("ID");
				rs.close();
				
				conn.commit();
				return lots;
			}
			catch (SQLException e)
			{	try
				{	conn.rollback();
					numTries++;
				}
				catch (SQLException se)
				{	throw new ServiceUnavailableException("Could not rollback when getting lots.");
				}
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not get lots.");	
	}

	/**
     * Returns the top level number of the car's current parking lot
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the system from returning the current lot's highest level
     * @return The top level number of the car's current parking lot
     */
	public int getMaxLevel() throws NotInLotException, ServiceUnavailableException 
	{
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{
			try
			{	// If we are not in a lot, then throw an exception
				if (lot == -1)
					throw new NotInLotException();

				PreparedStatement maxLevelStmt = conn.prepareStatement("SELECT MAX(Level) FROM Level WHERE LotID = ?");
				maxLevelStmt.setInt(1, this.lot);
				ResultSet rs = maxLevelStmt.executeQuery();
				rs.next();
				int maxLevel = rs.getInt(1);
				rs.close();
				return maxLevel;
			}
			catch (SQLException e)
			{	numTries++;
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not get max level.");
	}

	/**
     * Returns the bottom level number of the car's current parking lot
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the system from returning the current lot's lowest level
     * @return The bottom level number of the car's current parking lot
     */
	public int getMinLevel() throws NotInLotException, ServiceUnavailableException 
	{
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{
			try
			{	// If we are not in a lot, then throw an exception
				if (lot == -1)
					throw new NotInLotException();
				
				PreparedStatement minLevelStmt = conn.prepareStatement("SELECT MIN(Level) FROM Level WHERE LotID = ?");
				minLevelStmt.setInt(1, this.lot);
				ResultSet rs = minLevelStmt.executeQuery();
				rs.next();
				int minLevel = rs.getInt(1);
				rs.close();
				return minLevel;
			}
			catch (SQLException e)
			{	numTries++;
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not get min level.");
	}

	/**
     * Returns an array of other lots that have availability, sorted by lot distance such that
     * closer lots are listed first
     * @param lot The lot to get lot distances from
     * @throws LotNotFoundException Thrown if the given lot number is not known to the system
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if
     * some other reason prevents the system from discovering the availability of other lots
     * @return An array of other lots that have availability
     */
	public int[] getOtherLotAvailability(int lot) throws LotNotFoundException, ServiceUnavailableException 
	{
		printLotsStatus("When getting other lot availability", seq);
		
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{
			try
			{	PreparedStatement validLot = conn.prepareStatement("SELECT 'X' FROM Lot WHERE ID = ?");
				validLot.setInt(1, lot);
				ResultSet rs = validLot.executeQuery();
				
				if (!rs.next()) {
					rs.close();
					conn.rollback();
					throw new LotNotFoundException(lot);
				}
				rs.close();
				
				PreparedStatement otherLots = conn.prepareStatement("SELECT ToLotID FROM LotDistance D "
						+ "INNER JOIN Lot L ON L.ID=D.ToLotID WHERE FromLotID = ? AND CarCount < Capacity "
						+ "ORDER BY Distance");
				otherLots.setInt(1, lot);
				rs = otherLots.executeQuery();
				
				rs.last();
				int[] otherAvailableLots = new int[rs.getRow()];
				rs.beforeFirst();
				
				int i = 0;
				while (rs.next())
					otherAvailableLots[i++] = rs.getInt("ToLotID");
				rs.close();
				
				return otherAvailableLots;
			}
			catch (SQLException e)
			{	numTries++;
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not get other lot availability.");
	}

	/**
     * Moves the car from its present level to the level beneath it
     * @param seq The latest sequence number of the client
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws AtTopLevelException Thrown if the car is already on the lowest level
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some other reason prevents the car from moving to the lower level
     * @return The level number that the client's car is now on
     */
	public int moveDownLevel(int seq) throws NotInLotException, AtBottomLevelException, ServiceUnavailableException
	{
		printLevelsStatus("Before moving down one level", seq);
		
		// Check if request already completed
		if (alreadyCompleted(seq))
		{	System.out.println("Request already completed (in move down level)");
			printLevelsStatus("After moving down one level", seq);
			return level;
		}
		
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{
			try
			{	// If we are not in a lot, then throw an exception
				if (lot == -1)
					throw new NotInLotException();

				// If we are on the bottom level, then throw an exception
				if (level == 1)
					throw new AtBottomLevelException(this.lot, this.level);
				
				// Get car count of level
				PreparedStatement levelCarCount = conn.prepareStatement("SELECT CarCount FROM Level WHERE LotID=? AND Level=?");
				levelCarCount.setInt(1, lot);
				levelCarCount.setInt(2, level);
				ResultSet rs = levelCarCount.executeQuery();
				rs.next();
				int count = rs.getInt("CarCount");
				rs.close();
				
				// If lot car count is not zero, decrement count
				if (count > 0)
				{	PreparedStatement exitLevel = conn.prepareStatement(
						"UPDATE Level SET CarCount = CarCount - 1 WHERE LotID=? AND Level=?");
				    exitLevel.setInt(1, lot);
				    exitLevel.setInt(2, level);
				    exitLevel.execute();
				}
				
				// Decrement the level, but don't update the cached value yet
				int currLevel = level;
				currLevel--;
				
				// Get car count of level
				levelCarCount.setInt(2, currLevel);
				rs = levelCarCount.executeQuery();
				rs.next();
				count = rs.getInt("CarCount");
				rs.close();
				
				// Increment the level count
				PreparedStatement enterLevel = conn.prepareStatement(
						"UPDATE Level SET CarCount = CarCount + 1 WHERE LotID=? AND Level=?");
				enterLevel.setInt(1, lot);
				enterLevel.setInt(2, currLevel);
				enterLevel.execute();
				
				printLevelsStatus("After moving down one level", seq);
				
				// Update client data and commit the transaction
				updateClientData(seq, this.lot, currLevel);
				conn.commit();
				
				// With changes committed, update our cached values
				this.seq = seq;
				level = currLevel;
				
				return level;
			}
			catch (SQLException e)
			{	try
				{	conn.rollback();
					numTries++;
				}
				catch (SQLException se)
				{	throw new ServiceUnavailableException("Could not rollback when moving down level.");
				}
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not move down level.");	
	}

	/**
     * Moves the car from its present level to the level above it
     * @param seq The latest sequence number of the client
     * @throws NotInLotException Thrown if the car is not in a lot
     * @throws AtTopLevelException Thrown if the car is already on the highest level
     * @throws ServiceUnavailableException Thrown if the database cannot be contacted or if some
     * other reason prevents the car from moving to the next highest level
     * @return The level number that the client's car is now on
     */
	public int moveUpLevel(int seq) throws NotInLotException, AtTopLevelException, ServiceUnavailableException
	{
		printLevelsStatus("Before moving up one level", seq);
		
		// Check if request already completed
		if (alreadyCompleted(seq))
		{	System.out.println("Request already completed (in move up level)");
			printLevelsStatus("After moving up one level", seq);
			return level;
		}
		
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{
			try
			{   
				// If we are not in a lot, then throw an exception
				if (lot == -1)
					throw new NotInLotException();

				// If we are on the top level, then throw an exception
				PreparedStatement levelQuery = conn.prepareStatement("SELECT MAX(Level) HighestLevel FROM Level WHERE LotID=?");
				levelQuery.setInt(1, lot);
				ResultSet rs = levelQuery.executeQuery();
				rs.next();
				int maxLevel = rs.getInt("HighestLevel");
				if (level == maxLevel) {
					rs.close();
					conn.rollback();
					throw new AtTopLevelException(this.lot, this.level);
				}
				rs.close();
				
				// Get car count of level
				PreparedStatement levelCarCount = conn.prepareStatement("SELECT CarCount FROM Level WHERE LotID=? AND Level=?");
				levelCarCount.setInt(1, lot);
				levelCarCount.setInt(2, level);
				rs = levelCarCount.executeQuery();
				rs.next();
				int count = rs.getInt("CarCount");
				rs.close();
				
				// If lot car count is not zero, decrement count
				if (count > 0)
				{	PreparedStatement exitLevel = conn.prepareStatement(
						"UPDATE Level SET CarCount = CarCount - 1 WHERE LotID=? AND Level=?");
				    exitLevel.setInt(1, lot);
				    exitLevel.setInt(2, level);
				    exitLevel.execute();
				}
				
				// Increment the level, but don't update the cached value yet
				int currLevel = level;
				currLevel++;
				
				// Get car count of level
				levelCarCount.setInt(2, currLevel);
				rs = levelCarCount.executeQuery();
				rs.next();
				count = rs.getInt("CarCount");
				rs.close();
				
				// Increment the level count
				PreparedStatement enterLevel = conn.prepareStatement(
						"UPDATE Level SET CarCount = CarCount + 1 WHERE LotID=? AND Level=?");
				enterLevel.setInt(1, lot);
				enterLevel.setInt(2, currLevel);
				enterLevel.execute();
				
				printLevelsStatus("After moving up one level", seq);
				
				// Update client data and commit the transaction
				updateClientData(seq, this.lot, currLevel);
				conn.commit();
				
				// With changes committed, update our cached values
				this.seq = seq;
				level = currLevel;
				
				return level;
			}
			catch (SQLException e)
			{	try
				{	conn.rollback();
					numTries++;
				}
				catch (SQLException se)
				{	throw new ServiceUnavailableException("Could not rollback when moving up level.");
				}
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not move up level.");	
	}
	
	/**
	 * Checks if the current client request has already been completed.
	 * @param seq
	 * @return
	 * @throws ServiceUnavailableException
	 */
	private boolean alreadyCompleted(int seq) throws ServiceUnavailableException
	{
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{	try
			{	this.checkDuplicate.setInt(1, this.clientID);
				this.checkDuplicate.setInt(2, seq);
				ResultSet rs = this.checkDuplicate.executeQuery();
				this.checkDuplicate.clearParameters();
				
				// Check if new request
				boolean completed = rs.next();
				rs.close();
				return completed;
			}
			catch (SQLException e)
			{	numTries++;
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not check if request was already completed.");
	}
	
	/**
	 * Updates the client table, which is used for duplicate detection.
	 * @param seq
	 * @param lot
	 * @param level
	 * @return
	 * @throws ServiceUnavailableException
	 */
	private void updateClientData(int seq, int lot, int level) throws ServiceUnavailableException
	{
		int numTries = 0;
		while (numTries < MAX_ATTEMPTS)
		{	try
			{	this.updateSequence.setInt(1, seq);
				
				if (lot == -1)
					this.updateSequence.setNull(2, Types.INTEGER);
				else
					this.updateSequence.setInt(2, lot);
				
				if (level == -1)
					this.updateSequence.setNull(3, Types.INTEGER);
				else
					this.updateSequence.setInt(3, level);
				
				this.updateSequence.setInt(4, this.clientID);
				
				if (this.updateSequence.executeUpdate() != 1) 
				{	conn.rollback();
					throw new ServiceUnavailableException("Update on Client table failed unexpectedly");
				}
				
				this.updateSequence.clearParameters();
				return;
			}
			catch (SQLException e)
			{	numTries++;
			}
		}
		
		// If max attempts exceeded, throw exception
		System.err.println("Failed to execute commands on database, even after " + MAX_ATTEMPTS + " attempts.");
		throw new ServiceUnavailableException("Could not check if request was already completed.");
	}
	
	/**
	 * Print the status of all the lots
	 * @param message The message to display with the status
	 * @param seq The client request sequence number
	 */
	private void printLotsStatus(String message, int seq)
	{
		if (PRINT_LOTS_STATUS)
		{	try
			{	PreparedStatement lotsStatus = conn.prepareStatement("SELECT ID, CarCount FROM Lot");
				ResultSet rs = lotsStatus.executeQuery();
				
				System.out.println("=== Lots Status === (" + message + ") for Client " + this.clientID + " Seq " + seq);
				while (rs.next())
					System.out.println("Lot: " + rs.getInt("ID") + "\tCarCount: " + rs.getInt("CarCount"));
				rs.close();
				System.out.println();
			}
			catch (SQLException e)
			{	System.err.println("Could not print lot status.");
			}
		}
	}
	
	/**
	 * Print the status of all the levels in the current lot
	 * @param message The message to display with the status
	 * @param seq The client request sequence number
	 */
	private void printLevelsStatus(String message, int seq)
	{
		if (PRINT_LEVELS_STATUS)
		{	try
			{	PreparedStatement levelsStatus = conn.prepareStatement("SELECT Level, CarCount FROM Level WHERE LotID = ?");
				levelsStatus.setInt(1, this.lot);
				ResultSet rs = levelsStatus.executeQuery();
				
				System.out.println("=== Levels Status === (" + message + ") for Client " + this.clientID + " Seq " + seq);
				while (rs.next())
					System.out.println("Level: " + rs.getInt("Level") + "\tCarCount: " + rs.getInt("CarCount"));
				rs.close();
				System.out.println();
			}
			catch (SQLException e)
			{	System.err.println("Could not print levels status.");
			}
		}
	}
}
