package ParkNPark.middletier;

import java.io.PrintStream;
import java.util.ArrayList;

import org.omg.CORBA.ORB;
import org.omg.CORBA.SystemException;
import org.omg.CosNaming.Binding;
import org.omg.CosNaming.BindingHolder;
import org.omg.CosNaming.BindingIterator;
import org.omg.CosNaming.BindingIteratorHolder;
import org.omg.CosNaming.BindingListHolder;
import org.omg.CosNaming.NameComponent;
import org.omg.CosNaming.NamingContext;
import org.omg.CosNaming.NamingContextExt;
import org.omg.CosNaming.NamingContextExtHelper;
import org.omg.CosNaming.NamingContextHelper;
import org.omg.CosNaming.NamingContextPackage.AlreadyBound;
import org.omg.CosNaming.NamingContextPackage.CannotProceed;
import org.omg.CosNaming.NamingContextPackage.InvalidName;
import org.omg.CosNaming.NamingContextPackage.NotFound;
import org.omg.PortableServer.POA;
import org.omg.PortableServer.POAHelper;

import ParkNPark.common.InputEater;
import ParkNPark.interfaces.ClientManagerFactory;
import ParkNPark.interfaces.ClientManagerFactoryHelper;
import ParkNPark.interfaces.ReplicationManagerHelper;
import ParkNPark.interfaces.ReplicationManagerPOA;
import ParkNPark.interfaces.ServiceUnavailableException;

public class ReplicationManager extends ReplicationManagerPOA
{
	/** List of active servers */
	private ArrayList<ServerState> servers;
	
	/** Naming context */
	private NamingContextExt nameServer;
	
	/** Our naming service's ParkNPark context, which contains server registrations */
	private NamingContext parkNParkContext = null;
	
	/** Object request broker */
	private ORB orb;
	
	/** Root POA */
	private POA rootPOA;
	
	/**
	 * Entry point of replication manager
	 * @param args
	 */
	public static void main(String[] args)
	{
		ReplicationManager rm = new ReplicationManager();
		rm.servers = new ArrayList<ServerState>();
		
		// Set up replication manager objects
		rm.setupRMObjects(args);
		
		// Refresh the ParkNPark naming context and, if successful, initialize the
		// server list from that context
		if (!rm.refreshParkNParkContext())
			System.err.println("Could not refresh the ParkNPark naming context");
		else
			rm.initializeServerListFromParkNParkContext();
		
		// Set the active server to the last server
		System.out.println("Activating the last registered server");
		rm.setPrimary(0);
			
		// Notify the user that the server is now running
        System.out.println("Server running.");
		
// TODO: May need to create separate thread for this so that serverRegistered can be called by other processes (??)
        // From Steve: serverRegistered() can be called by other processes already. CORBA is running in
        // a background thread at this point (and the implementation works in practice)
		while (true)
			rm.checkServers();
	}
	
	/**
	 * Called by servers when they start up.
	 * @param ipAddress
	 * @param serviceName
	 * @param cmf
	 */
	public synchronized void serverRegistered(String ipAddress, String serviceName, ClientManagerFactory cmf) 
	{
		servers.add(new ServerState(ipAddress, serviceName, cmf));
		System.out.println("Server at " + ipAddress + " [ " + serviceName + "] is now active.");
		
		// If this is the only active server, then set as primary
		if (servers.size() == 1)
			setPrimary(0);
	}

	/** Shared static instance of the input eater output stream wrapped in a PrintStream */
	protected static PrintStream inputEater = new PrintStream(new InputEater());
	
	/**
	 * Check whether or not servers are active
	 */
	public void checkServers() 
	{
		boolean serverFailure = false;
		for (int i = 0; i < servers.size(); i++)
		{	ServerState s = servers.get(i);
			
			// Check to see if server is active
			PrintStream err = System.err;
			try
			{	
				
				// Tries to connect to the database, if fails 3 times throws an exception and the server is restarted
				
				System.setErr(inputEater);
				s.cmf.poke();
				System.setErr(err);
				
			}
			catch (Exception e)
			{	// Remove from active servers list
				System.setErr(err);
				servers.remove(i);
				System.out.println("Server at " + s.ipAddress + " [ " + s.serviceName + "] has failed.");
				
				// If this server is the primary, then set a new primary
				if (s.isPrimary)
					setPrimary(i);
				
				// Remove service name from naming service
				// TODO: If the server is manually started and calls serverRegistered()
				// between the line above and the line below, we might unregister a
				// possibly healthy server. How can we fix this (I haven't thought through it yet)?
				// (access to the servers List should probably be synchronized as ArrayList per-se
				// is not thread-safe..)
				
				// FIXME: the TODO above can be triggered by starting a server and, after it's ready,
				// immediately ctrl+c-ing it and restarting it. Several tries might be necessary,
				// but the effect is the message: "Could not unbind service go-172.19.132.119 due to exception org.omg.CosNaming.NamingContextPackage.NotFound"
				// (it tried to unregister it twice)
				unbindServer(s);
				
				// TODO: may need to create new thread to start server if it will take time (??)
				// Restart the failed server
				if (e instanceof ServiceUnavailableException)
					startServer(s, true);
				else
					startServer(s, false);
				
				// We had a server failure, so we should not go to sleep before the next call to
				// checkServers()
				serverFailure = true;
			}
		}
		
		// Sleep before our returning if we did not encounter any server failures in this call
		if (!serverFailure) {
			try {
				Thread.sleep(5000);  // 5 seconds
			} catch (InterruptedException e) { }
		}
	}
	
	/**
	 * Set one of the active servers as the primary
	 * @param index
	 */
	public void setPrimary(int index) 
	{
		ServerState s;
		
		// If only one server, make it primary
		if (servers.size() == 1)
		{	s = servers.get(0);
		}
		else if (servers.size() == 0)
		{
			System.out.println("No servers are active; notifying name service that no server is primary");

			// Register selected server as the primary in the naming service
			try
			{	NameComponent[] serverName = nameServer.to_name("ParkNParkPrimary"); 
			    parkNParkContext.unbind(serverName);
			} catch (NotFound e) {
				System.out.println("No primary server was registered");
			}
			catch (Exception e) {
				System.err.println("Could not remove the primary server reference from the name service due to exception: " + e.getClass().getName());
			}
			return;
		}
		else
		{	// Set primary server using most-recently-active mechanism
			if (index > 0)
				s = servers.get(index - 1);
			else
				s = servers.get(servers.size() - 1);
		}
			
		
		// If we don't have our ParkNPark naming context, then get it from the server now.
		if (parkNParkContext == null && !refreshParkNParkContext())
		{	System.err.println("Could not get ParkNPark context");
			return;
		}
		
		//		 Set selected server as the primary
		System.out.println("Setting server at " + s.ipAddress + " [ " + s.serviceName + "] as the primary.");
		
		// Register selected server as the primary in the naming service
		try
		{	NameComponent[] serverName = nameServer.to_name("ParkNParkPrimary"); 
			parkNParkContext.rebind(serverName, ClientManagerFactoryHelper.narrow(s.cmf));
			s.isPrimary = true;
		}
		catch (Exception e)
		{	System.err.println("Could not set server at " + s.ipAddress + " as the primary due to exception: " + e.getClass().getName());
// TODO: need backup plan if this fails; perhaps call setPrimary with i - 1 (??)		
		}
	}
	
	/**
	 * Unbind the failed server's service name from the naming service
	 * @param s
	 */
	public void unbindServer(ServerState s)
	{
		// If we don't have our ParkNPark naming context, then get it from the server now.
		if (parkNParkContext == null && !refreshParkNParkContext())
		{	System.err.println("Could not get ParkNPark context");
			return;
		}
		
		// Unbind the server from the naming service
		try
		{	NameComponent[] serverName = nameServer.to_name(s.serviceName); 
			parkNParkContext.unbind(serverName);
		}
		catch (Exception e)
		{	System.err.println("Could not unbind service " + s.serviceName + " due to exception " + e.getClass().getName());
		}
	}
	
	/**
	 * Start a new server or restart a corrupted server
	 * @param s
	 * @param corrupted
	 */
	public void startServer(ServerState s, boolean corrupted) 
	{
		// If corrupted, then first need to kill the server
		if (corrupted)
		{	System.out.println("Server at " + s.ipAddress + " [ " + s.serviceName + "] is having database connectivity problems.");
		
			// Kill the server
// TODO: implement logic to kill the server
		}
		
		// Restart the server
		System.out.println("Restarting server at " + s.ipAddress + " [ " + s.serviceName + "].");
// TODO: implement logic to start server
	}
	
	/**
	 * Setup CORBA objects
	 */
	private void setupRMObjects(String args[])
	{
		// Initialize the ORB
        System.out.println("Starting the CORBA ORB");
        try 
        {	orb = ORB.init(args, null);
        } 
        catch (Exception e) 
        {	System.out.println("Exception while creating ORB...terminating");
        	System.exit(1);
        }

        // Get the root POA and activate it
        PrintStream err = System.err;
        try 
        {
        	System.setErr(inputEater);
        	System.out.println("Activating the CORBA root POA");
            rootPOA = POAHelper.narrow(orb.resolve_initial_references("RootPOA"));
            rootPOA.the_POAManager().activate();
            System.setErr(err);
        } 
        catch (Exception e) 
        {
            System.setErr(err);
        	System.out.println("Exception while creating ORB; perhaps the port is in use due to another server instance on this server? Terminating");
        	System.exit(1);
        	return;
        }

        try 
        {	// Get the naming service object
            System.out.println("Fetching the name service reference");
            nameServer = NamingContextExtHelper.narrow(orb.resolve_initial_references("NameService"));
        } 
        catch (Exception e) 
        {	System.out.println("Cannot connect to the name service; please ensure that it is running and try again");
        	System.exit(1);
        	return;
        }
                
        try 
        {	// Get the proper context object for ParkNPark
            System.out.println("Fetching the ParkNPark name service context");
            NameComponent parkNParkContextName[] = nameServer.to_name("ParkNPark");
            NameComponent parkNParkReplicationManager[] = nameServer.to_name("ParkNParkReplicationManager");
            NamingContext parkNParkContext = null;
            do {
                try {
                    parkNParkContext = NamingContextHelper.narrow(nameServer.resolve(parkNParkContextName));
                } catch (NotFound e)
                {
                 	// Register the context
                  	try {
                   		System.out.println("ParkNPark name service context not found; creating");
                       	nameServer.bind_new_context(parkNParkContextName);
                   	} catch (AlreadyBound f) {
                   		System.out.println("Naming context already created! Trying the fetch again");
                   	}
                }
            } while (parkNParkContext == null);
            
            // Register the replication manager servant
            // Register client manager factory with the naming service
            System.out.println("Registering the client manager factory name with the CORBA naming service");
            try {
                parkNParkContext.bind(parkNParkReplicationManager, ReplicationManagerHelper.narrow(rootPOA.servant_to_reference(this)));
            } catch (AlreadyBound e) {
            	System.out.println("Server already registered! Overwriting the old name registration with myself");
                parkNParkContext.rebind(parkNParkReplicationManager, ReplicationManagerHelper.narrow(rootPOA.servant_to_reference(this)));
            }
        }
        catch(Exception e)
        {
        	
        }
	}
	
	/**
	 * Refreshes the parkNParkContext name service object that we use to get server
	 * bindings on. The name service is contacted from scratch in this method
	 * @return True if the parkNParkContext was refreshed or false if it was not
	 */
	private boolean refreshParkNParkContext()
	{	
		// Fetch the root naming context
		PrintStream err = System.err;
		try {
			System.setErr(inputEater);
			nameServer = NamingContextExtHelper.narrow(orb.resolve_initial_references("NameService"));
		    System.setErr(err);
		} 
		catch (Exception e) 
		{
		    System.setErr(err);
			System.err.println("Could not contact the naming service; please ensure that it is running.");
			parkNParkContext = null;
			return false;
		}

		// Get the proper context object for ParkNPark
		System.out.println("Fetching the ParkNPark name service context");
		NameComponent parkNParkContextName[];
		
		try 
		{	parkNParkContextName = nameServer.to_name("ParkNPark");
		} 
		catch (Exception e) 
		{	System.err.println("CORBA problem while communicating with the name server");
			parkNParkContext = null;
			return false;
		}
		
		try 
		{	parkNParkContext = NamingContextHelper.narrow(nameServer.resolve(parkNParkContextName));
		} 
		catch (NotFound e) 
		{	System.err.println("ParkNPark name service context not found; is the ParkNPark server up?");
			parkNParkContext = null;
			return false;
		}
		catch (Exception e) 
		{	System.err.println("CORBA problem while communicating with the name server");
			parkNParkContext = null;
			return false;
		}
		
		// We were able to get it :-)
		return true;
	}
	
	/**
	 * Initializes the server List from the ParkNParkContext in the name service.
	 * As a prerequisite, the caller must ensure that parkNParkContext is not null
	 */
	protected void initializeServerListFromParkNParkContext()
	{
		// Get the servers that are currently registered into a new set
		try {
			BindingListHolder listHolder = new BindingListHolder();
			BindingIteratorHolder listIteratorHolder = new BindingIteratorHolder();
			
			// See if the parkNParkContext is alive. If it's not, and we cannot refresh it,
			// then return false now
			try {
				System.out.println("Fetching the server list from the naming service");
			    parkNParkContext.list(0, listHolder, listIteratorHolder);
			} catch (Exception e)
			{
				// Refresh the context
				System.out.println("Could not get the server list from the naming service; replication manager will start with an empty server list");
				return;
			}
			
			// Iterate through the server list
			BindingIterator iterator = listIteratorHolder.value;
			BindingHolder currentBindingHolder = new BindingHolder();
			Binding currentBinding;
			ClientManagerFactory currentServer;
			while (iterator.next_one(currentBindingHolder)) {
				currentBinding = currentBindingHolder.value;
				
				// Narrow this server down to an object and add that server object reference
				if (currentBinding.binding_name != null) {
					try {
						// Only care about ClientManagerFactory registrations
						org.omg.CORBA.Object registeredObject = parkNParkContext.resolve(currentBinding.binding_name);
						if (registeredObject instanceof ClientManagerFactory) {
						    currentServer = ClientManagerFactoryHelper.narrow(registeredObject);
						    
						    // Get the name of the binding, which contains the IP address
						    String serviceName = nameServer.to_string(currentBinding.binding_name);
						    
					    	// Remove backslashes from the name (inserted by the name service)
						    serviceName = serviceName.replace("\\", "");
					    	
						    // Extract the IP address if this is a server registration with an IP address
						    int ipStart = serviceName.lastIndexOf('-');
						    if (ipStart > 0 && ipStart < serviceName.length() - 1) {
						    	String ipAddress = serviceName.substring(ipStart + 1);
						    	
						    	// Add this server registration
						    	servers.add(new ServerState(ipAddress, serviceName, currentServer));
								System.out.println("Server at " + ipAddress + " [ " + serviceName + "] is possibly active.");
						    }
						}
					} catch (NotFound e) {
						System.out.println("A server's name registration went away during the server registration fetch; ignoring that server");
					} catch (InvalidName e) {
						System.out.println("A server's name registration somehow became syntactically invalid; ignoring that server");
					} catch (CannotProceed e) {
						System.out.println("A server's name registration could not be retrieved due to an unspecified reason (CannotProceed exception); ignoring that server");
					}
				}
			}
		} catch (SystemException e) {
			System.out.println("CORBA problem while fetching the server set from the name server: " + e.getClass().getName());
		}
	}
}
