package ParkNPark.middletier;

import java.io.PrintStream;
import java.net.InetAddress;
import java.net.UnknownHostException;

import org.omg.CORBA.ORB;
import org.omg.CORBA.SystemException;
import org.omg.CosNaming.NameComponent;
import org.omg.CosNaming.NamingContext;
import org.omg.CosNaming.NamingContextExt;
import org.omg.CosNaming.NamingContextExtHelper;
import org.omg.CosNaming.NamingContextHelper;
import org.omg.CosNaming.NamingContextPackage.AlreadyBound;
import org.omg.CosNaming.NamingContextPackage.NotFound;
import org.omg.PortableServer.POA;
import org.omg.PortableServer.POAHelper;

import ParkNPark.common.InputEater;
import ParkNPark.interfaces.ClientManagerFactoryHelper;
import ParkNPark.interfaces.ReplicationManagerHelper;
import ParkNPark.interfaces.ReplicationManager;

/**
 * Starts the server from scratch by activating CORBA, registering our client
 * manager factory with the naming service, and handing control over to CORBA
 */
public class Server
{
	/**
	 * Starts the ParkNPark server
	 * @param args
	 */
	public static void main(String[] args)
	{
        NamingContextExt nameServer;
        ORB orb = null;
        NameComponent serverName[], parkNParkContextName[];
	    
        // Initialize the ORB
        System.out.println("Starting the CORBA ORB");
        try {
            orb = ORB.init(args, null);
        } catch (Exception e) {
        	System.out.println("Exception while activating the root POA; perhaps your JVM is malfunctioning");
        	System.exit(1);
        	return;
        }

        // Get the root POA and activate it
        POA rootPOA;
        PrintStream err = System.err, inputEater = new PrintStream(new InputEater());
        try {
        	System.setErr(inputEater);
            System.out.println("Activating the CORBA root POA");
            rootPOA = POAHelper.narrow(orb.resolve_initial_references("RootPOA"));
            rootPOA.the_POAManager().activate();
            System.setErr(err);
        } catch (Exception e) {
            System.setErr(err);
        	System.out.println("Exception while creating ORB; perhaps the port is in use due to another server instance on this server?");
        	System.exit(1);
        	return;
        }
       
        System.out.println("Fetching the name service reference");
        try {
            // Get the naming service object
        	System.setErr(inputEater);
            nameServer = NamingContextExtHelper.narrow(orb.resolve_initial_references("NameService"));
            System.setErr(err);
        } catch (Exception e) {
            System.setErr(err);
        	System.out.println("Cannot connect to the name service; please ensure that it is running and try again");
        	System.exit(1);
        	return;
        }
                
        // Create our client manager factory servant object
        System.out.println("Creating the client manager factory");
        ClientManagerFactoryImpl clientManagerFactory = new ClientManagerFactoryImpl();

        try {
            // Get the proper context object for ParkNPark
            System.out.println("Fetching the ParkNPark name service context");
            parkNParkContextName = nameServer.to_name("ParkNPark");
            NamingContext parkNParkContext = null;
            do {
                try {
                    parkNParkContext = NamingContextHelper.narrow(nameServer.resolve(parkNParkContextName));
                } catch (NotFound e)
                {
                 	// Register the context
                  	try {
                   		System.out.println("ParkNPark name service context not found; creating");
                       	nameServer.bind_new_context(parkNParkContextName);
                   	} catch (AlreadyBound f) {
                   		System.out.println("Naming context already created! Trying the fetch again");
                   	}
                }
            } while (parkNParkContext == null);
            
            // Get the replication manager instance
            ReplicationManager replicationManager;
            try {
            	NameComponent replicationManagerName[] = nameServer.to_name("ParkNParkReplicationManager");
            	replicationManager = ReplicationManagerHelper.narrow(parkNParkContext.resolve(replicationManagerName));
            } catch (NotFound e) {
            	System.out.println("Warning: Could not find the replication manager");
            	replicationManager = null;
            }
            
            // Get the server's primary IP address and host name
            String ipAddr, name;
            try {
                InetAddress addr = InetAddress.getLocalHost();
                ipAddr = addr.getHostAddress();
                String hostname = addr.getHostName();
                name = hostname + "-" + ipAddr;
                serverName = nameServer.to_name(name);
                System.out.println("Registering server as " + name);
            } catch (UnknownHostException e) {
               	System.out.println("Could not get IP address! Exiting");
               	System.exit(1);
               	return;
            }
            
            // Register client manager factory with the naming service
            System.out.println("Registering the client manager factory name with the CORBA naming service");
            try {
                parkNParkContext.bind(serverName, ClientManagerFactoryHelper.narrow(rootPOA.servant_to_reference(clientManagerFactory)));
            } catch (AlreadyBound e) {
            	System.out.println("Server already registered! Overwriting the old name registration with myself");
                parkNParkContext.rebind(serverName, ClientManagerFactoryHelper.narrow(rootPOA.servant_to_reference(clientManagerFactory)));
            }

            // Attempt to register with the replication manager, not caring if it does not work
            // because the replication manager can pick us up from the name service later
            if (replicationManager != null) {
                err = System.err;
                System.out.println("Notifying the replication manager that we are now active");
                try {
                	System.setErr(inputEater);
                    replicationManager.serverRegistered(ipAddr, name, ClientManagerFactoryHelper.narrow(rootPOA.servant_to_reference(clientManagerFactory)));
                    System.setErr(err);
                } catch (SystemException e) {
                    System.setErr(err);
                	System.out.println("Warning: CORBA communication problem while trying to notify the replication manager that we are up: " + e.getClass().getName());
                } catch (Exception e) {
                    System.setErr(err);
                    e.printStackTrace();
                    System.exit(1);
                    return;
                }
            }
        } catch (Exception e) {
        	System.out.println("Could not register with the name service due to an exception " + e.getClass().getName() + "; exiting");
        	e.printStackTrace();
           	System.exit(1);
           	return;
        }
        
        // Wait for invocations from clients
        System.out.println("Server running.");
        orb.run();
    }
}
